/*
 *  KnotTTY is a remote secure shell using the new public key exchange protocol
 *  with the Braid Group Cryptosystems
 *
 *  Copyright (C) 2006 Adrien Pinet
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 */

#include "CSessionFTP.h"

#include "CBuffer.h"
#include "CSHADigest.h"
#include "CProgression.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <signal.h>

CSessionFTP::CSessionFTP()
{

}

CSessionFTP::~CSessionFTP()
{

}

int CSessionFTP::Init(char ip[], int port, char userName[], char password[])
{
	CBuffer SessionMod(1);
	CBuffer Status(1);

	SessionMod.GetBuffer()[0] = FTP;

	if(!SecureConnection.ConnectToServer(ip, port))
	return 0;

	if(!SecureConnection.SecureConnection())
	return 0;
	
	if(!SecureConnection.SecureSend((uByte*)userName, strlen(userName) + 1))
	return 0;

	if(!SecureConnection.SecureSend((uByte*)password, strlen(password) + 1))
	return 0;

	if(!SecureConnection.SecureSend(SessionMod.GetBuffer(), SessionMod.GetMaxBufferSize()))
	return 0;

	if(!SecureConnection.SecureReceive(Status.GetBuffer(), Status.GetMaxBufferSize(), &Status.bufferSize))
	return 0;

	return Status.GetBuffer()[0] == AUTHENTICATED;
}

int CSessionFTP::Upload(char srcFilePathName[], char destFilePathName[])
{
	CBuffer Data(CSecureConnection::MAXSOCKETDATASIZE);
	uByte cmd = UPLOAD;
	uInt32 fileSize;

	if(!SecureConnection.SecureSend(&cmd, 1))
	return 0;

	if(!SecureConnection.SecureSend((uByte*) destFilePathName, strlen(destFilePathName) + 1))
	return 0;

	if(!GetFileSize(srcFilePathName, &fileSize))
	return 0;
	
	if(!SecureConnection.SecureSend((uByte*) &fileSize, sizeof(fileSize)))
	return 0;

	FILE* pFile = fopen(srcFilePathName, "r");

	if(pFile == NULL)
	return 0;

	uInt32 numBlock = fileSize / Data.GetMaxBufferSize();
	uInt32 lastBlockSize = fileSize % Data.GetMaxBufferSize();

	CProgression Progression;

	if(lastBlockSize)
	Progression.Init(numBlock + 1, 1, srcFilePathName);
	else
	Progression.Init(numBlock, 1, srcFilePathName);


	CSHADigest SHADigest;
	SHADigest.Start();

	for(uInt32 i = 0 ; i < numBlock ; i++)
	{
		Progression.Draw();

		fread(Data.GetBuffer(), sizeof(uByte), Data.GetMaxBufferSize(), pFile);
		SHADigest.ComputeInputBuffer(Data.GetBuffer(), Data.GetMaxBufferSize());
		
		if(!SecureConnection.SecureSend(Data.GetBuffer(), Data.GetMaxBufferSize()))
		{
			fclose(pFile);
			return 0;
		}

		Progression.StepIt();
	}

	if(lastBlockSize)
	{
		Progression.Draw();

		fread(Data.GetBuffer(), sizeof(uByte), lastBlockSize, pFile);
		SHADigest.ComputeInputBuffer(Data.GetBuffer(), lastBlockSize);

		if(!SecureConnection.SecureSend(Data.GetBuffer(), lastBlockSize))
		{
			fclose(pFile);
			return 0;
		}

		Progression.StepIt();
	}
	
	SHADigest.Stop();
	fclose(pFile);

	if(!SecureConnection.SecureSend(SHADigest.GetByteResult(), 20))
	return 0;

	return 1;
}

int CSessionFTP::Download(char srcFilePathName[], char destFilePathName[])
{
	CBuffer Data(CSecureConnection::MAXSOCKETBUFFERSIZE);
	CBuffer FileSize(sizeof(uInt32));
	CBuffer SHADigestResult(20);
	uByte cmd = DOWNLOAD;
	uInt32 fileSize;

	if(!SecureConnection.SecureSend(&cmd, 1))
	return 0;
	
	if(!SecureConnection.SecureSend((uByte*) srcFilePathName, strlen(srcFilePathName) + 1))
	return 0;

	if(!SecureConnection.SecureReceive((uByte*) &fileSize, FileSize.GetMaxBufferSize(), &FileSize.bufferSize))
	return 0;

	FILE* pFile = fopen(destFilePathName, "w");

	if(pFile == NULL)
	return 0;

	uInt32 numBlock = fileSize / Data.GetMaxBufferSize();
	uInt32 lastBlockSize = fileSize % Data.GetMaxBufferSize();

	CProgression Progression;

	if(lastBlockSize)
	Progression.Init(numBlock + 1, 1, srcFilePathName);
	else
	Progression.Init(numBlock, 1, srcFilePathName);

	CSHADigest SHADigest;
	SHADigest.Start();

	for(uInt32 i = 0 ; i < numBlock ; i++)
	{
		Progression.Draw();

		if(!SecureConnection.SecureReceive(Data.GetBuffer(), Data.GetMaxBufferSize(), &Data.bufferSize))
		{
			fclose(pFile);
			return 0;
		}

		fwrite(Data.GetBuffer(), sizeof(uByte), Data.GetBufferSize(), pFile);
		SHADigest.ComputeInputBuffer(Data.GetBuffer(), Data.GetBufferSize());

		Progression.StepIt();
	}

	if(lastBlockSize)
	{
		Progression.Draw();

		if(!SecureConnection.SecureReceive(Data.GetBuffer(), Data.GetMaxBufferSize(), &Data.bufferSize))
		{
			fclose(pFile);
			return 0;
		}

		fwrite(Data.GetBuffer(), sizeof(uByte), Data.GetBufferSize(), pFile);
		SHADigest.ComputeInputBuffer(Data.GetBuffer(), Data.GetBufferSize());

		Progression.StepIt();
	}
	
	SHADigest.Stop();
	fclose(pFile);

	if(!SecureConnection.SecureReceive(SHADigestResult.GetBuffer(), SHADigestResult.GetMaxBufferSize(), &SHADigestResult.bufferSize))
	return 0;

	for(uInt32 i = 0 ; i < 20 ; i++)
	{
		if(SHADigest.GetByteResult()[i] != SHADigestResult.GetBuffer(i))
		return 0;
	}

	return 1;
}

int CSessionFTP::SendLS()
{

	return 1;
}

int CSessionFTP::SendCD()
{

	return 1;
}


int CSessionFTP::GetFileSize(char fileName[], uInt32* pFileSize)
{
	FILE* pFile = fopen(fileName, "r");

	if(pFile == NULL)
	return 0;
	
	if(fseek(pFile, 0, SEEK_END))
	{
		fclose(pFile);
		return 0;
	}

	*pFileSize = ftell(pFile);

	fclose(pFile);

	return 1;
}

int CSessionFTP::GetFileName(char filePathName[], char fileName[])
{
	int filePathNameSize = strlen(filePathName);

	if(filePathNameSize == 0)
	return 0;

	int i = filePathNameSize - 1;
	
	while(i > 0 && filePathName[i] != '/')
	i--;

	if(i > 0)
	i++;

	for(int j = i ; j <= filePathNameSize ; j++)
	fileName[j - i] = filePathName[j];

	return 1;
}
