/*
 *  KnotTTY is a remote secure shell using the new public key exchange protocol
 *  with the Braid Group Cryptosystems
 *
 *  Copyright (C) 2006 Adrien Pinet
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 */

#include "CChannelManager.h"

#include <stdio.h>

template <class TConnection, class TEvent>
CChannelManager<TConnection, TEvent>::CChannelManager()
{
	ChannelList = NULL;
}

template <class TConnection, class TEvent>
CChannelManager<TConnection, TEvent>::~CChannelManager()
{
	StopChannelManager();

	if(ChannelList)
	{
		delete[] ChannelList;
		ChannelList = NULL;
	}
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::StartChannelManager(TConnection* pConnection, TEvent* pEvent, uInt32 maxChannel)
{
	this->pConnection = pConnection;
	this->pEvent = pEvent;
	this->maxChannel = maxChannel;

	if(ChannelList)
	delete[] ChannelList;

	ChannelList = new CChannel[maxChannel];

	pthread_attr_t thread_attr;

	pthread_attr_init(&thread_attr);
	pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);

	return pthread_create(&threadChannelManager, &thread_attr, ThreadChannelManager, this) == 0;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::StopChannelManager()
{
	if(!ChannelList)
	return 0;

	int ret = pthread_cancel(threadChannelManager);

	for(uInt32 idChannel = 0 ; idChannel < maxChannel ; idChannel++)
	{
		if(ChannelList[idChannel].IsOpen())
		ChannelList[idChannel].Close();
	}

	return ret == 0;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::Open(uInt32* pIdChannel)
{
	for(uInt32 idChannel = 0 ; idChannel < maxChannel ; idChannel++)
	{
		if(ChannelList[idChannel].IsClose())
		{
			if(!ChannelList[idChannel].Open())
			return 0;

			*pIdChannel = idChannel;
			
			if(SendOpen(idChannel))
			return 1;
			
			return 0;
		}
	}

	return 0;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::Close(uInt32 idChannel)
{
	if(idChannel >= maxChannel || ChannelList[idChannel].IsClose())
	return 0;

	if(!ChannelList[idChannel].Close())
	return 0;
	
	if(SendClose(idChannel))
	return 1;

	return 0;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::Write(uInt32 idChannel, uByte data[], uInt32 dataSize)
{
	if(idChannel >= maxChannel || ChannelList[idChannel].IsClose())
	return 0;

	uInt32 bufferSize = dataSize + 2;
	uByte* buffer = new uByte[bufferSize];
	
	buffer[0] = idChannel;
	buffer[1] = DATA;
	
	for(uInt32 i = 0 ; i < dataSize ; i++)
	buffer[2 + i] = data[i];
	
	if(!pConnection->SecureSend(buffer, bufferSize))
	{
		StopChannelManager();
		return 0;
	}

	return 1;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::Read(uInt32 idChannel, uByte data[], uInt32 maxDataSize, uInt32* pDataSize)
{
	if(idChannel >= maxChannel || ChannelList[idChannel].IsClose())
	return 0;

	if(!ChannelList[idChannel].Read(data, maxDataSize, pDataSize))
	return 0;
	
	return 1;
}



template <class TConnection, class TEvent>
void* CChannelManager<TConnection, TEvent>::ThreadChannelManager(void* pvCChannelManager)
{
	CChannelManager* pChannelManager = (CChannelManager*) pvCChannelManager;
	
	uByte* buffer = new uByte[pChannelManager->pConnection->MAXSOCKETDATASIZE];
	uInt32 bufferSize;

	while(pChannelManager->pConnection->SecureReceive(buffer, TConnection::MAXSOCKETDATASIZE, &bufferSize))
	{
		if(bufferSize < 2)
		{
			delete[] buffer;
			return NULL;
		}

		uInt32 idChannel = buffer[0];
		uInt32 flag = buffer[1];

		switch(flag)
		{
		case OPEN:
			if(pChannelManager->ChannelList[idChannel].Open())
			pChannelManager->pEvent->OnOpen(idChannel);
			break;

		case DATA:
			if(pChannelManager->ChannelList[idChannel].IsOpen())
			pChannelManager->ChannelList[idChannel].Write(buffer + 2, bufferSize - 2);
			else
			pChannelManager->SendClose(idChannel);
			break;

		case CLOSE:
			if(pChannelManager->ChannelList[idChannel].Close())
			pChannelManager->pEvent->OnClose(idChannel);
			break;
		}
	}

	delete[] buffer;
	pChannelManager->StopChannelManager();

	return NULL;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::SendOpen(uInt32 idChannel)
{
	uByte message[2];

	message[0] = idChannel;
	message[1] = OPEN;
	
	if(!pConnection->SecureSend(message, 2))
	{
		StopChannelManager();
		return 0;
	}

	return 1;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::SendClose(uInt32 idChannel)
{
	uByte message[2];

	message[0] = idChannel;
	message[1] = CLOSE;
	
	if(!pConnection->SecureSend(message, 2))
	{
		StopChannelManager();
		return 0;
	}

	return 1;
}


template <class TConnection, class TEvent>
CChannelManager<TConnection, TEvent>::CChannel::CChannel()
{
	data = NULL;
	dataSize = 0;
	isOpen = 0;
}

template <class TConnection, class TEvent>
CChannelManager<TConnection, TEvent>::CChannel::~CChannel()
{
	Close();
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::CChannel::Open()
{
	if(IsOpen())
	return 0;

	pthread_mutex_init(&mutexData, NULL);
	pthread_mutex_init(&mutexWait, NULL);
	pthread_cond_init(&condWait, NULL);

	isOpen = 1;

	return 1;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::CChannel::Close()
{
	if(IsClose())
	return 0;

	isOpen = 0;

	pthread_cond_signal(&condWait);
	FreeData();

	pthread_mutex_unlock(&mutexData);
	pthread_mutex_unlock(&mutexWait);

	pthread_mutex_destroy(&mutexData);
	pthread_mutex_destroy(&mutexWait);
	pthread_cond_destroy(&condWait);

	return 1;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::CChannel::IsOpen()
{
	return isOpen;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::CChannel::Write(uByte data[], uInt32 dataSize)
{
	if(IsClose())
	return 0;

	/* Wait that buffer is Empty */
	pthread_mutex_lock(&mutexWait);

	if(IsDataFull())
	pthread_cond_wait(&condWait, &mutexWait);

	if(!WriteData(data, dataSize))
	{
		pthread_mutex_unlock(&mutexWait);
		return 0;
	}

	pthread_cond_signal(&condWait);
	
	pthread_mutex_unlock(&mutexWait);

	return IsOpen();
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::CChannel::Read(uByte data[], uInt32 maxDataSize, uInt32* pDataSize)
{
	if(IsClose())
	return 0;

	/* Wait that buffer is full */
	pthread_mutex_lock(&mutexWait);
	
	if(IsDataEmpty())
	pthread_cond_wait(&condWait, &mutexWait);

	if(!ReadData(data, maxDataSize, pDataSize))
	{
		pthread_mutex_unlock(&mutexWait);
		return 0;
	}

	if(!FreeData())
	{
		pthread_mutex_unlock(&mutexWait);
		return 0;
	}
	
	pthread_cond_signal(&condWait);
	
	pthread_mutex_unlock(&mutexWait);

	return IsOpen();
}


template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::CChannel::IsDataFull()
{
	int ret = 1;
	pthread_mutex_lock(&mutexData);

	if(data == NULL)
	ret = 0;

	pthread_mutex_unlock(&mutexData);
	
	return ret;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::CChannel::WriteData(uByte data[], uInt32 dataSize)
{
	pthread_mutex_lock(&mutexData);
	
	if(this->data)
	{
		pthread_mutex_unlock(&mutexData);
		return 0;
	}

	this->data = new uByte[dataSize];
	this->dataSize = dataSize;

	for(uInt32 i = 0 ; i < dataSize ; i++)
	this->data[i] = data[i];

	pthread_mutex_unlock(&mutexData);
	return 1;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::CChannel::ReadData(uByte data[], uInt32 maxDataSize, uInt32* pDataSize)
{
	pthread_mutex_lock(&mutexData);
	
	if(this->data == NULL || maxDataSize < dataSize)
	{
		pthread_mutex_unlock(&mutexData);
		return 0;
	}

	for(uInt32 i = 0 ; i < dataSize ; i++)
	data[i] = this->data[i];

	*pDataSize = dataSize;

	pthread_mutex_unlock(&mutexData);
	return 1;
}

template <class TConnection, class TEvent>
int CChannelManager<TConnection, TEvent>::CChannel::FreeData()
{
	pthread_mutex_lock(&mutexData);
	
	if(data)
	{
		delete[] data;
		data = NULL;
		dataSize = 0;

		pthread_mutex_unlock(&mutexData);

		return 1;
	}

	pthread_mutex_unlock(&mutexData);

	return 0;
}
