/*
 *  KnotTTY is a remote secure shell using the new public key exchange protocol
 *  with the Braid Group Cryptosystems
 *
 *  Copyright (C) 2006 Adrien Pinet
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 */

#include "CSecureServer.h"

#include "CBraid.h"
#include "CSHADigest.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>

#define BRAIDSIZE 500

CSecureServer::CSecureConnection::CSecureConnection(int socketClient, char ipClient[])
{
	this->socketClient = socketClient;
	strcpy(this->ipClient, ipClient);

	pthread_mutex_init(&mutexToSend, NULL);
	pthread_mutex_init(&mutexToReceive, NULL);

	connectionSecured = 0;
}

CSecureServer::CSecureConnection::~CSecureConnection()
{
	Close();
}

int CSecureServer::CSecureConnection::SecureConnection()
{
	if(connectionSecured)
	return 1;

	CBraid Braid(BRAIDSIZE);
	uInt32 XBraid[BRAIDSIZE];
	uInt32 aBraid[BRAIDSIZE];
	uInt32 a1Braid[BRAIDSIZE];
	uInt32 aXa1Braid[BRAIDSIZE];
	uInt32 bXb1Braid[BRAIDSIZE];
	uInt32 tempBraid[BRAIDSIZE];
	uInt32 abXa1b1Braid[BRAIDSIZE];

	uByte sharedPassword[32];
	uInt32 sizeReceive;
	
	/* Generation random braids*/
	Braid.RandomLeftBraid(aBraid);
	Braid.RandomBraid(XBraid);

	/* Compute the first public key (X, aXa-1) */
	Braid.Reverse(aBraid, a1Braid);
	Braid.Product(aBraid, XBraid, tempBraid);
	Braid.Product(tempBraid, a1Braid, aXa1Braid);

	/* Sending the first public key (X, aXa-1) */
	if(!Send((uByte*) XBraid, sizeof(uInt32) * BRAIDSIZE)) return 0;
	if(!Send((uByte*) aXa1Braid, sizeof(uInt32) * BRAIDSIZE)) return 0;

	/* Receiving the second public key (bXb-1) */
	if(!Receive((uByte*)bXb1Braid, sizeof(uInt32) * BRAIDSIZE, &sizeReceive)) return 0;
	if(sizeReceive != sizeof(uInt32) * BRAIDSIZE) return 0;

	/* Computing the shared private key (abXa-1b-1) */
	Braid.Product(aBraid, bXb1Braid, tempBraid);
	Braid.Product(tempBraid, a1Braid, abXa1b1Braid);
	Braid.BraidToPassword(abXa1b1Braid, sharedPassword);

	/* Securing the connection */
	AesToSend.Init(sharedPassword);
	AesToReceive.Init(sharedPassword);

	/* Verifying the security */
	uByte tokenSended[20], tokenReceived[20];
	CSHADigest SHADigest;
	
	SHADigest.Start();
	SHADigest.ComputeInputBuffer(sharedPassword, 32);
	SHADigest.Stop();
	
	SHADigest.GetByteResult(tokenSended);
	
	if(!SecureSend(tokenSended, 20)) return 0;
	if(!SecureReceive(tokenReceived, 20, &sizeReceive)) return 0;
	
	for(uInt32 i = 0 ; i < 20 ; i++)
	{
		if(tokenReceived[i] != tokenSended[i])
		return 0;
	}
	
	connectionSecured = 1;

	return 1;
}

void CSecureServer::CSecureConnection::Close()
{
	connectionSecured = 0;

	shutdown(socketClient, SHUT_RDWR);
	close(socketClient);
}

int CSecureServer::CSecureConnection::SecureSend(uByte data[], uInt32 dataSize)
{
	LockToSend();

	if(dataSize > MAXSOCKETDATASIZE)
	{
		UnLockToSend();
		return 0;
	}
	
	uByte checkSum[2];
	uInt32 bufferSize = dataSize + 2;
	uByte* buffer = new uByte[bufferSize];
	uByte* encryptedBuffer = new uByte[bufferSize];
	
	ComputeCheckSum(data, dataSize, checkSum);
	
	buffer[0] = checkSum[0];
	buffer[1] = checkSum[1];
	
	for(uInt32 i = 0 ; i < dataSize ; i++)
	buffer[i + 2] = data[i];
	
	AesToSend.Encrypt(buffer, encryptedBuffer, bufferSize);
	
	int retSend = Send(encryptedBuffer, bufferSize);
	
	delete[] buffer;
	delete[] encryptedBuffer;

	UnLockToSend();
	return retSend;
}

int CSecureServer::CSecureConnection::SecureReceive(uByte data[], uInt32 maxDataSize, uInt32* pDataSize)
{
	LockToReceive();

	uByte* encryptedBuffer = new uByte[MAXSOCKETBUFFERSIZE];
	uInt32 bufferSize;
	
	if(!Receive(encryptedBuffer, MAXSOCKETBUFFERSIZE, &bufferSize))
	{
		delete[] encryptedBuffer;
		UnLockToReceive();
		return 0;
	}

	uByte checkSum[2];
	uByte* buffer = new uByte[bufferSize];
	AesToReceive.Decrypt(encryptedBuffer, buffer, bufferSize);

	uInt32 dataSize = bufferSize - 2;
	
	if(dataSize > maxDataSize)
	{
		delete[] buffer;
		delete[] encryptedBuffer;
		UnLockToReceive();
		return 0;
	}
	
	checkSum[0] = buffer[0];
	checkSum[1] = buffer[1];
	
	for(uInt32 i = 0 ; i < dataSize ; i++)
	data[i] = buffer[i + 2];
	
	delete[] buffer;
	delete[] encryptedBuffer;

	*pDataSize = dataSize;
	
	UnLockToReceive();

	return VerifyCheckSum(data, dataSize, checkSum);
}

char* CSecureServer::CSecureConnection::GetIp()
{
	return ipClient;
}

int CSecureServer::CSecureConnection::Send(uByte buffer[], uInt32 bufferSize)
{
	if(bufferSize > MAXSOCKETBUFFERSIZE)
	return 0;

	/* Building trame */
	uInt32 trameSize = 4 + bufferSize;
	uByte* trame = new uByte[trameSize];

	uByte checkSum[2];
	ComputeCheckSum(buffer, bufferSize, checkSum);
	
	trame[0] = (uByte) bufferSize;
	trame[1] = (uByte) (bufferSize >> 8);
	trame[2] = checkSum[0];
	trame[3] = checkSum[1];

	for(uInt32 i = 0 ; i < bufferSize ; i++)
	trame[i + 4] = buffer[i];

	/* Sending trame */
	uInt32 trameSizeSend = 0;

	while(trameSizeSend < trameSize)
	{
		int currentTrameSizeSend = write(socketClient, trame + trameSizeSend, trameSize - trameSizeSend);
		
		if(currentTrameSizeSend <= 0)
		{
			delete[] trame;
			return 0;
		}
	
		trameSizeSend += currentTrameSizeSend;
	}

	delete[] trame;
	return 1;
}

int CSecureServer::CSecureConnection::Receive(uByte buffer[], uInt32 maxBufferSize, uInt32* pBufferSize)
{
	/* Receiving header */
	uByte header[4];
	uInt32 headerSize = 4;
	uInt32 headerSizeReceive = 0;
	
	while(headerSizeReceive < headerSize)
	{
		int currentHeaderSizeReceive = read(socketClient, header + headerSizeReceive, headerSize - headerSizeReceive);
		
		if(currentHeaderSizeReceive <= 0)
		return 0;
		
		headerSizeReceive += currentHeaderSizeReceive;
	}
	
	/* Reading header */
	uByte checkSum[2];
	uInt32 bufferSize = header[1];
	bufferSize = (bufferSize << 8) | header[0];

	checkSum[0] = header[2];
	checkSum[1] = header[3];

	if(bufferSize > maxBufferSize)
	return 0;
	
	/* Receiving buffer */
	uInt32 bufferSizeReceive = 0;

	
	while(bufferSizeReceive < bufferSize)
	{
		int currentBufferSizeReceive = read(socketClient, buffer + bufferSizeReceive, bufferSize - bufferSizeReceive);

		if(currentBufferSizeReceive <= 0)
		return 0;

		bufferSizeReceive += currentBufferSizeReceive;
	}
	
	*pBufferSize = bufferSize;
	
	return VerifyCheckSum(buffer, bufferSize, checkSum);
}

void CSecureServer::CSecureConnection::LockToSend()
{
	pthread_mutex_lock(&mutexToSend);
}

void CSecureServer::CSecureConnection::UnLockToSend()
{
	pthread_mutex_unlock(&mutexToSend);
}

void CSecureServer::CSecureConnection::LockToReceive()
{
	pthread_mutex_lock(&mutexToReceive);
}

void CSecureServer::CSecureConnection::UnLockToReceive()
{
	pthread_mutex_unlock(&mutexToReceive);
}

void CSecureServer::CSecureConnection::ComputeCheckSum(uByte buffer[], uInt32 bufferSize, uByte checkSum[])
{
	checkSum[0] = 0x00;
	checkSum[1] = 0xFF;

	for(uInt32 i = 0 ; i < bufferSize ; i++)
	checkSum[i % 2] ^= buffer[i];
}

int CSecureServer::CSecureConnection::VerifyCheckSum(uByte buffer[], uInt32 bufferSize, uByte checkSum[])
{
	uByte oldCheckSum[2];
	ComputeCheckSum(buffer, bufferSize, oldCheckSum);

	return oldCheckSum[0] == checkSum[0] && oldCheckSum[1] == checkSum[1];
}

CSecureServer::CSecureServer()
{
}

CSecureServer::~CSecureServer()
{
	CloseServer();
}

int CSecureServer::CreateServer(int portServer, int maxConnection)
{
	Error = NOERROR;
	signal(SIGPIPE, SIG_IGN);
	
	this->maxConnection = maxConnection;
	this->portServer = portServer;
	socketServer = 0;
	
	addrinfo *res = NULL;
	addrinfo addrinfoServer;

	memset(&addrinfoServer, 0x00, sizeof(addrinfoServer));

	addrinfoServer.ai_family = AF_INET;
	addrinfoServer.ai_flags = AI_PASSIVE;
	addrinfoServer.ai_socktype = SOCK_STREAM;

	char sPortServer[6];
	sprintf(sPortServer, "%d", portServer);

	if (getaddrinfo(NULL, sPortServer, &addrinfoServer, &res) < 0)
	{
		Error |= CREATESERVERERROR;
		return 0;
	}

	 /* Create the listen socket */
	 socketServer = socket(res->ai_family, res->ai_socktype, res->ai_protocol);

	if(socketServer < 0 || bind(socketServer, res->ai_addr, res->ai_addrlen) < 0 || listen(socketServer, maxConnection) < 0)
	{
		Error |= CREATESERVERERROR;
		freeaddrinfo(res);
		return 0;
	}

	freeaddrinfo(res);
	return 1;
}

void CSecureServer::CloseServer()
{
	shutdown(socketServer, SHUT_RDWR);
	close(socketServer);
}

CSecureServer::CSecureConnection* CSecureServer::WaitConnection()
{
	int socketClient;
	char ipClient[16];
	
	sockaddr_in address;
	socklen_t adressSize = sizeof(sockaddr_in);

	socketClient = accept(socketServer, (sockaddr*)&address, &adressSize);

	if(socketClient < 0)
	return NULL;

	sprintf(ipClient, "%d.%d.%d.%d", (address.sin_addr.s_addr >> 0) & 0xFF,
						(address.sin_addr.s_addr >> 8) & 0xFF,
						(address.sin_addr.s_addr >> 16) & 0xFF,
						(address.sin_addr.s_addr >> 24) & 0xFF);

	return new CSecureConnection(socketClient, ipClient);
}

uInt32 CSecureServer::GetError(void)
{
	return Error;
}
